! ESA (C) 2000-2018 
!
! This file is part of ESA's XMM-Newton Scientific Analysis System (SAS).
!
!    SAS is free software: you can redistribute it and/or modify
!    it under the terms of the GNU General Public License as published by
!    the Free Software Foundation, either version 3 of the License, or
!    (at your option) any later version.
!
!    SAS is distributed in the hope that it will be useful,
!    but WITHOUT ANY WARRANTY; without even the implied warranty of
!    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
!    GNU General Public License for more details.
!
!    You should have received a copy of the GNU General Public License
!    along with SAS.  If not, see <http://www.gnu.org/licenses/>.
!
!+
! Module: rotdet2sky_mod
!
! Authors:  Brendan Perry - NASA/GSFC (comprehensive rewrite, Spring 2020)
!           Steven L. Snowden - NASA/GSFC original algorithm
!           
!
! Description: (former) Extended Source Analysis Software
!
! Comments:
!
! $Id: rotdet2sky_mod.f90, v 1.3 2009/02/26 16:39:45 aibarra Exp $
!-
!  Version 5.0  9 November 2012
!  Version 0.1 30 March 2020 (including rename rot_det_sky to rotdet2sky)
!   [BHP - new version removed from ESAS package into this standalone package]
!
!  WAS: ESAS task rot_det_sky 
!  Rotates images in detector coordinates into sky coordinates provided
!  by a template image.  This was used in the XMM-ESAS package to rotate
!  the model particle background images, which are created by the programs
!  mosback/pnback/proton in detector coordinates, into sky coordinates. This
!  new version is modular (not only for ESAS uses). Support was added for
!  swcx images. 
!
!  INPUT:
!    Input FOV image (template) (900,900), e.g. mos1S001-fovimc.fits
!    Input background image (780,780), e.g. mos1S001-
!    Input mask image (optional) (900,900)
!    makeoutname (boolean) Don't choose outfilename?
!    Output rotated backround image (900,900) (user selected name)
!    DETX central coord
!    DETY central coord
!    X reference pixel
!    Y reference pixel
!    Mask file name (missing or "none" defaults to no masking,
!        same size as the template)
!
! NEW example: rotdet2sky infovimage=mos1S001-fovimt-350-1100.fits \
!              inbkgimage=mos1S001-protimdet-350-1100.fits \
!              outbkgimage=mos1S001-protimsky-350-1100.fits \
!              withmask=F withskyxy=T detx=detxfromesky2det dety=detyfromesky2det \
!              skyx=RAPNTfromImage   skyy=DECPNTfromImage
!
!    OLD example: rot-back 1 1S001 350 1250 -1406.1 -1368.6 \
!                 450.9 450.9 mask.fits
!
!    Template image file name (hardwired to 900x900 pixels)
!    Detector image file name (hardwired to 780x780 pixels)
!
!  OUTPUT:
!    Tranformed image image (same size as the template)

module rotdet2sky_mod

  use dal
  use param 
  use errorHandling
  use sasdal
  use gofutils_mod  ! from goflib, formerly in marq_mod
  use gofmatrix_mod ! from goflib, formerly in marq_mod
  use gofio_mod     ! from goflib, formerly in subsn (SS)
  
  implicit none

  private

  character(len=*), parameter :: &
       infovimpar  = "infovimage", &  ! template image file
       inbkgimpar  = "inbkgimage", &  ! input bkg image file to be rotated
       outbkgimpar = "outbkgimage", & ! output rotated bkg image file 
       withmaskpar = "withmask", &    ! provide a mask file?
       maskfilepar = "inmaskimage", & ! input optional mask image
       detxcenpar  = "detx", &        ! input detx (from esky2det)
       detycenpar  = "dety", &        ! input dety (from esky2det)
       withskypar  = "withskyxy", &   ! input skyx/y or use crpix1/2?
       skyxpar     = "skyx", &        ! user input skyx/y if user
       skyypar     = "skyy"           ! chooses withskyxy=T (default)
  
  public rotdet2sky

contains

  !---------------------------------------------------------------------
  subroutine rotdet2sky
    !---------------------------------------------------------------------
    
    type(DataSetT) :: inFOVIMG_S, inBKGIMG_S, inMASKIMG_S, outBKGIMG_S
    type(ArrayT)   :: inFOVIMG_A, inBKGIMG_A, inMASKIMG_A, outBKGIMG_A
    type(TableT)   :: in_T
    
    character(len=5) :: expido, expidb, expidm, insto, instb, instm, celow, cehigh
    character(len=8) :: ctypen(2), ctypeo(2), cunito(2), &
         ctypeb(2), cunitb(2), ctypem(2), cunitm(2), suffix
    character(len=10) :: obsido, obsidb, obsidm, prefix
    character(len=12) :: estring
    character(len=25) :: smodeo, smodeb, smodem, filtero, filterb, filterm, &
         radecsyso, radecsysb, radecsysm, type
    character(len=60) :: com
    character(len=256) :: inFOVIMG_F, inBKGIMG_F, inMASKIMG_F, outBKGIMG_F
    integer :: elowo, elowb, elowm, ehigho, ehighb, ehighm
    integer :: i, ii, ix, iy, j, jj, lun, n, frmtime
    integer :: err = 0, typeF, nRowso, nRowsb, nRowsm
    integer :: revoluto, revolutb, revolutm
    integer :: status = 0, xSizeo, ySizeo, xSizem, ySizem, xSizeb, ySizeb
    integer :: outBKGIMGdim(2)
    integer, parameter :: xGood=780, yGood=780
    integer(kind=int16), dimension(:,:), pointer :: aDI16
    integer(kind=int32), dimension(:,:), pointer :: inFOVIMG_IP, inMASKIMG_IP, &
         inBKGIMG_IP, aDI32
    logical :: warndie, withmask, withskyxy, i_ok, withlong
    real(kind=single) :: backo, expoo, areao, backb, expob, areab, backm, expom, aream
    real(kind=double) :: crvalo(2), crpixo(2), cdelto(2), crotao(2), &
         crvalb(2), crpixb(2), cdeltb(2), crotab(2), crvalm(2), crpixm(2), &
         cdeltm(2), crotam(2), crotan, rapnto, decpnto, papnto, rapntb, &
         decpntb, papntb, rapntm, decpntm, papntm
    real(kind=double) :: dxcen, dycen
    real(kind=single) :: alpha, beta, input(780,780), rotarr(900,900), &
         mask(900,900), maskit(900,900), rn, skyx, skyy, theta, x, y
    real(double), dimension(:,:), pointer :: aDR64
    real(single), dimension(:,:), pointer :: inFOVIMG_RP, inBKGIMG_RP, inMASKIMG_RP, &
         outBKGIMG_RP
    real(single), dimension(:,:), pointer :: aDR32
    withmask    = .FALSE.
    withskyxy   = .FALSE.
    withlong    = .TRUE. ! check this!!!???
    i_ok        = .FALSE.
    skyx = 0.
    skyy = 0. 
    inFOVIMG_F  = stringParameter(infovimpar)
    inBKGIMG_F  = stringParameter(inbkgimpar)
    withmask    = booleanParameter(withmaskpar)
    inMASKIMG_F = ""
    if (withmask) then
       inMASKIMG_F = stringParameter(maskfilepar)
    endif
    dxcen     = realParameter(detxcenpar)
    dycen     = realParameter(detycenpar)
    !
    ! open, read in TEMPLATE image (900,900) and its coordinate keywords
    !
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(inFOVIMG_F),inFOVIMG_S,1,status,warndie)
    status = 0
    warndie = .TRUE.
    typeF = 3 ! 1 = EVLI, 2 = SPEC, 3 = IMG
    call read_header_info(typeF,inFOVIMG_S,in_T,inFOVIMG_A,insto,obsido,& 
         expido,smodeo,frmtime,filtero,revoluto,backo,areao,expoo,&
         elowo,ehigho,nRowso,warndie)
    warndie = .TRUE.
    ! get coordinate keywords from template image 
    call coord_keys (inFOVIMG_S,rapnto,decpnto,papnto,crvalo,cdelto,&
         cunito,ctypeo,crpixo,crotao,radecsyso,warndie)
    ! get template image from, e.g. mos1S001-fovimt.fits (900,900,R32 array)???!!!
    call read_2Darray(inFOVIMG_S,inFOVIMG_A,inFOVIMG_RP,aDR64,aDI16,aDI32,xSizeo,ySizeo)
    withskyxy = booleanParameter(withskypar)
    if (withskyxy) then 
       skyx  = realParameter(skyxpar)
       skyy  = realParameter(skyypar)
    else
       skyx = crpixo(1)
       skyy = crpixo(2)
    endif
    ! why???!!!
    if(skyx .lt. 100.0) skyx = 100.0*skyx                                              
    if(skyy .lt. 100.0) skyy = 100.0*skyy     
    write(celow,"(I5)") elowo
    write(cehigh,"(I5)") ehigho
    estring = "-" // trim(adjustl(celow)) // "-" // trim(adjustl(cehigh))
    !
    select case (insto(1:2))
    case ('M1','M2')
       prefix = "mos" // trim(adjustl(insto(2:2))) // trim(adjustl(expido))
    case ('PN')
       prefix = "pn" // trim(adjustl(expido))
    case default
       call error("INSTwrong",&
            "Instrument must be M1, M2, or PN only, check obj input file")
    end select
    !
    if (withmask) then 
       write(errstr,*) "Optional Mask image: ",trim(adjustl(inMASKIMG_F))
       call message(AppMsg,Always)
       ! open, read in MASK image and its coordinate keywords
       status = 0
       warndie = .TRUE.
       call check_open_set(trim(inMASKIMG_F),inMASKIMG_S,1,status,warndie)
       status = 0
       warndie = .TRUE.
       typeF = 3 ! 1 = EVLI, 2 = SPEC, 3 = IMG
       call read_header_info(typeF,inMASKIMG_S,in_T,inMASKIMG_A,instm,obsidm,& 
            expidm,smodem,frmtime,filterm,revolutm,backm,aream,expom,elowm,ehighm,&
            nRowsm,warndie)
       warndie = .TRUE.
       call coord_keys (inMASKIMG_S,rapntm,decpntm,papntm,crvalm,cdeltm,&
            cunitm,ctypem,crpixm,crotam,radecsysm,warndie)
       call read_2Darray(inMASKIMG_S,inMASKIMG_A,inMASKIMG_RP,aDR64,aDI16,inMASKIMG_IP,&
            xSizem,ySizem)
    else
       write(errstr,*) "No optional Mask Image selected."
       call message(AppMsg,Always)
    endif
    ! user-specified, but if still 'default', generate filename.
    ! this works for running rotdet2sky from the command line. In
    ! proton and swcx, rotdet2sky will automatically be called with
    ! outBKGIMG_F set to something correct. I.e. if run standalone,
    ! it will make mos1S001-bkgimsky...fits from keywords, but prot
    ! will create mos1S001-protimsky...fits and swcx will create
    ! mos1S001-swcximsky.fits when they run rotdet2sky within those
    ! tasks. Put this in docs???!!!
    !
    outBKGIMG_F = stringParameter(outbkgimpar)
    if (trim(adjustl(outBKGIMG_F)) .eq. 'default') then
       outBKGIMG_F = trim(adjustl(prefix)) // "-bkgimsky" // &
            trim(adjustl(estring)) // ".fits" ! ???!!! what if no elow/ehigh???
    end if
    write(errstr,*) "Output BKG IMAGE (SKY coords) = " // trim(adjustl(outBKGIMG_F))
    call message(AppMsg,Always)
    !
    ! open, read INPUT background image file and its coordinate keywords
    !
    status = 0
    warndie = .TRUE.
    call check_open_set(trim(inBKGIMG_F),inBKGIMG_S,1,status,warndie)
    status = 0
    warndie = .TRUE.
    typeF = 3 ! 1 = EVLI, 2 = SPEC, 3 = IMG
    call read_header_info(typeF,inBKGIMG_S,in_T,inBKGIMG_A,instb,obsidb,& 
         expidb,smodeb,frmtime,filterb,revolutb,backb,areab,expob,elowb,&
         ehighb,nRowsb,warndie)
    warndie = .TRUE.
    call coord_keys (inBKGIMG_S,rapntb,decpntb,papntb,crvalb,cdeltb,&
         cunitb,ctypeb,crpixb,crotab,radecsysb,warndie)
    call read_2Darray(inBKGIMG_S,inBKGIMG_A,inBKGIMG_RP,aDR64,aDI16,inBKGIMG_IP,&
         xSizeb,ySizeb)
    !
    ! end parameter input section    
    !
    n = 0
    rn = 0.
    ! print out params
    write(errstr,'(A,1X,A)')'Template image for casting DET to SKY: ',trim(adjustl(inFOVIMG_F))
    call message(AppMsg,Always)
    write(errstr,'(A,A,1X,A)')'Template image coordinate system   (CTYPE): ',&
         trim(adjustl(ctypeo(1))),trim(adjustl(ctypeo(2)))
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Template image central coordinates (CRVAL): ',&
         crvalo(1)," ",crvalo(2)
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Template image reference pixel     (CRPIX): ',&
         crpixo(1)," ",crpixo(2)
    call message(AppMsg,Always)
    write(errstr,'(A,E10.4,A,E10.4)')'Template image pixel sizes         (CDELT): ',&
         cdelto(1)," ",cdelto(2)
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3)')'Template image rotation angle      (CROTN): ',&
         papnto
    !    write(errstr,'(A,F8.3,A,F8.3)') 'Template FOV image rotation angle      (CROTN): ',&
    !         crotao(1)," ",crotao(2)
    call message(AppMsg,Always)
    write(errstr,'(A,I5,A,I5)')'Template image nRows               (NAXIS): ',&
         xSizeo," ",ySizeo
    call message(AppMsg,Always)

    ! hmmm, typical input files don't have EXPOSURE (should they???!!!) keywords,
    ! and EXPOSURE isn't actually USED anywhere below, sooooo, remove this???!!!
    ! write(errstr,*) 'Template image EXPOSURE     : ',expoo
    ! call message(AppMsg,Noisy)
    !
    write(errstr,'(A,1X,A)')'Input BKG image to be cast from DET to SKY: ',&
         trim(adjustl(inBKGIMG_F))
    call message(AppMsg,Always)
    write(errstr,'(A,A,1X,A)')'Input BKG image coordinate system   (CTYPE): ',&
         trim(adjustl(ctypeb(1))),trim(adjustl(ctypeb(2)))
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Input BKG image central coordinates (CRVAL): ',&
         crvalb(1)," ",crvalb(2)
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Input BKG image reference pixel     (CRPIX): ',&
         crpixb(1)," ",crpixb(2)
    call message(AppMsg,Always)
    write(errstr,'(A,E10.4,A,E10.4)')'Input BKG image pixel sizes         (CDELT): ',&
         cdeltb(1)," ",cdeltb(2)
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3)')'Input BKG image rotation angle      (CROTN): ',&
         papntb
    !    write(errstr,'(A,F8.3,A,F8.3)') 'OV image rotation angle      (CROTN): ',&
    !         crotab(1)," ",crotab(2)
    call message(AppMsg,Always)
    write(errstr,'(A,I5,A,I5)')'Input BKG image nRows               (NAXIS): ',&
         xSizeb," ",ySizeb
    call message(AppMsg,Always)
    !
    if (withmask) then
       write(errstr,'(A,1X,A)')'Optional MASK to be applied: ',&
            trim(adjustl(inMASKIMG_F))
       call message(AppMsg,Always)
       write(errstr,'(A,A,1X,A)')'MASK image coordinate system   (CTYPE): ',&
            trim(adjustl(ctypem(1))),trim(adjustl(ctypem(2)))
       call message(AppMsg,Always)
       write(errstr,'(A,F8.3,A,F8.3)')'MASK image central coordinates (CRVAL): ',&
            crvalm(1)," ",crvalm(2)
       call message(AppMsg,Always)
       write(errstr,'(A,F8.3,A,F8.3)')'MASK image reference pixel     (CRPIX): ',&
            crpixm(1)," ",crpixm(2)
       call message(AppMsg,Always)
       write(errstr,'(A,E10.4,A,E10.4)')'MASK image pixel sizes         (CDELT): ',&
            cdeltm(1)," ",cdeltm(2)
       call message(AppMsg,Always)
       write(errstr,'(A,F8.3)')'MASK image rotation angle      (CROTN): ',&
            papntm
       !    write(errstr,'(A,F8.3,A,F8.3)') 'Template FOV image rotation angle      (CROTN): ',&
       !         crotam(1)," ",crotam(2)
       call message(AppMsg,Always)
       write(errstr,'(A,I5,A,I5)')'MASK image nRows               (NAXIS): ',&
            xSizeo," ",ySizeo
       call message(AppMsg,Always)
    end if
    ! why CROTN = PA_PNT in old rot_det_sky???!!! 
    crotan = papnto ! mos1 only

    if (insto(1:2) .eq. "M2" ) then 
       crotan = papnto + 90.0  
    elseif (insto(1:2) .eq. "PN" ) then ! PN 
       crotan = papnto - 90.0
    endif

    ctypeo = ctypeb
    cdelto = cdeltb
    dxcen = (dxcen + 19499.0)/50.0 + 0.5                                           
    dycen = (dycen + 19499.0)/50.0 + 0.5      
    crpixo(1) = (-dxcen + 19499.0)/50.0 - 0.5 
    crpixo(2) = ( dycen + 19499.0)/50.0 + 0.5

    ! modes eliminated in SAS-19, any input goes, fixed with params
    ! mode = 4, 5 also eliminated in SAS-19 (Fall, 2020), mode 4 taken care of in SAS
    
    ! Cast the new observation into the mosaic

    theta = -0.01745329 * crotan

    alpha = skyx + dxcen * cos(theta) - dycen * sin(theta)
    beta  = skyy - dycen * cos(theta) - dxcen * sin(theta)

    x = alpha - dxcen * cos(theta) + dycen * sin(theta)
    y = beta  + dycen * cos(theta) + dxcen * sin(theta)
    ! this...doesn't do anything in the example???!!!
    do i=1,xGood ! 780
       do ii=1,yGood ! 780
           if(inFOVIMG_RP(i-1,ii-1) .ne. 0.) then  ! was: input(780,780)
             x = alpha - i * cos(theta) + ii * sin(theta)
             y = beta + ii * cos(theta) + i * sin(theta)
             ix = nint(x)
             iy = nint(y)
             if((ix .ge. 1) .and. (ix .le. 900) .and. &
                  (iy .ge. 1) .and. (iy .le. 900)) then
                rotarr(ix,iy) = rotarr(ix,iy) + inFOVIMG_RP(i,ii)
                n = n + 1
                rn = rn + inFOVIMG_RP(i,ii)
             endif
          else
          endif
       enddo
    enddo
    write(errstr,*) 'Non-zero input pixels = ',n
    call message(AppMsg,Always)
    write(errstr,*) 'Cast counts           = ',rn
    call message(AppMsg,Always)

    !  Do the standard masking

    if (withmask) then 
       n = 0
       rn = 0.0
       do i=1,900
          do ii=1,900
             if(inMASKIMG_RP(i-1,ii-1) .eq. 0) then
                rotarr(i,ii) = 0.
             else
                n = n + 1
                rn = rn + rotarr(i,ii)
             endif
          enddo
       enddo
       !
       write(errstr,*) 'Values after standard masking:'
       call message(AppMsg,Always)
       write(errstr,*) 'Non-zero input pixels = ',n
       call message(AppMsg,Always)
       write(errstr,*) 'Cast counts           = ',rn
    endif
    !  Print the details of the output image
    write(errstr,'(A,1X,A)')'Final (rotated) image in SKY coords: ',&
         trim(adjustl(outBKGIMG_F))
    call message(AppMsg,Always)
    write(errstr,'(A,A,1X,A)')'Final SKY image coordinate system   (CTYPE): ',&
         trim(adjustl(ctypem(1))),trim(adjustl(ctypem(2)))
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Final SKY image central coordinates (CRVAL): ',&
         crvalm(1)," ",crvalm(2)
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3,A,F8.3)')'Final SKY image reference pixel     (CRPIX): ',&
         crpixm(1)," ",crpixm(2)
    call message(AppMsg,Always)
    write(errstr,'(A,E10.4,A,E10.4)')'Final SKY image pixel sizes         (CDELT): ',&
         cdeltm(1)," ",cdeltm(2)
    call message(AppMsg,Always)
    write(errstr,'(A,F8.3)')'Final SKY image rotation angle      (CROTN): ',&
         papntm
    !    write(errstr,'(A,F8.3,A,F8.3)') 'Template FOV image rotation angle      (CROTN): ',&
    !         crotam(1)," ",crotam(2)
    call message(AppMsg,Always)
    write(errstr,'(A,I5,A,I5)')'Final SKY image nRows               (NAXIS): ',&
         xSizeo," ",ySizeo
    call message(AppMsg,Always)
    !
    ! open output image file outBKGIMG_F (unless clobber)
    !
    inquire(FILE=trim(outBKGIMG_F), EXIST=i_ok)
    if (i_ok) then
       if ( clobber() ) then ! checks SAS_CLOBBER env variable
          write(errstr,*) "Output BKGIMG file ",trim(adjustl(outBKGIMG_F)),&
               " exists and CLOBBER is set, will overwrite existing file."
          call message(AppMsg,Always)          
          call unlink (trim(outBKGIMG_F))
       else
          write(errstr,*)"Output BKGIMG file ",trim(adjustl(outBKGIMG_F)),&
               " exists and CLOBBER is NOT set...exiting"
          call error("cannotCLOBBER",trim(errstr))
       endif
    else
       write(errstr,*) "Output BKGIMG file ",trim(adjustl(outBKGIMG_F)),&
            " does not exist, cloning Input BKGIMG file."
       call message(AppMsg,Verbose)  
    end if
    ! open output background map
    ! here we clone the template image (900,900) 
    write(errstr,*) "Cloning Template IMAGE..."
    call message(AppMsg,Always)  
    outBKGIMG_S = clone(trim(adjustl(inFOVIMG_F)),trim(adjustl(outBKGIMG_F)),MODIFY)
    ! then we work on IT, recasting it's array values to rotarr from above
    call read_2Darray(outBKGIMG_S,outBKGIMG_A,outBKGIMG_RP,aDR64,aDI16,aDI32,&
         xSizeo,ySizeo)
    write(errstr,*) "Filling new output BKG IMAGE array with rotated values..."
    call message(AppMsg,Always)
    print*,'xsize, ysize = ',xSizeo," ",ySizeo

    do i=0,xSizeo-1
       do j=0,ySizeo-1
          outBKGIMG_RP(i,j) = rotarr(i+1,j+1)
       end do
    end do

    ! clean these keywords up and add HISTORY/COMMENT lines!!!
    !    call copyAllAttributes(attributable(inBKGIMG_A),attributable(outBKGIMG_A))
    !    call copyAllAttributes(attributable(inBKGIMG_S),attributable(outBKGIMG_S))
    !
    write(errstr,*) "Releasing (closing) output BKG IMAGE file."
    call message(AppMsg,Always)
    
    call release(outBKGIMG_S)
    return

  end subroutine rotdet2sky

end module rotdet2sky_mod
